# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Make these names available from nncf.common.graph directly
from nncf.common.graph.graph import LayerName as LayerName
from nncf.common.graph.graph import NNCFGraph as NNCFGraph
from nncf.common.graph.graph import NNCFGraphEdge as NNCFGraphEdge
from nncf.common.graph.graph import NNCFGraphPatternIO as NNCFGraphPatternIO
from nncf.common.graph.graph import NNCFNode as NNCFNode
from nncf.common.graph.graph import NNCFNodeName as NNCFNodeName
