# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nncf.tensor.definitions import TensorBackend as TensorBackend
from nncf.tensor.definitions import TensorDataType as TensorDataType
from nncf.tensor.definitions import TensorDeviceType as TensorDeviceType
from nncf.tensor.tensor import Tensor as Tensor
from nncf.tensor.tensor import unwrap_tensor_data as unwrap_tensor_data
